/*
 * Copyright 2007 Etienne Studer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nu.studer.idea.notifier;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Iterator;

import com.intellij.ide.plugins.RepositoryHelper;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.updateSettings.impl.PluginDownloader;
import com.intellij.openapi.util.BuildNumber;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.extensions.PluginId;
import com.intellij.openapi.util.JDOMUtil;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
// boaz was:
//import com.intellij.util.ObjectUtils;
//boaz now:
import org.apache.commons.lang.ObjectUtils;

import com.intellij.util.net.HttpConfigurable;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.filter.ElementFilter;

/**
 * This class queries the IDEA plugin repository and extracts all plugins that can be updated for a given IDEA build number.
 *
 * @author <a href="mailto:intellij@studer.nu">Etienne Studer</a>, Nov 7, 2006
 */
public class PluginRepositoryManager {
    private static final Logger LOG = Logger.getInstance(PluginRepositoryManager.class.getName());

    @NonNls
    private static final String TEST_MODE_PROPERTY_KEY = "pluginUpdateNotifier.testMode";

    @NonNls
    private static final String INTELLIJ_NET_PLUGINS_LIST = "http://plugins.intellij.net/plugins/list/?build=";

    @NonNls
    public static final String INTELLIJ_NET_PLUGIN_DOWNLOAD =
            "http://plugins.intellij.net/pluginManager/?action=download&id=PLUGIN_ID&build=BUILD";

    @NonNls
    private static final String IDEA_PLUGIN_ELEMENT = "idea-plugin";
    @NonNls
    private static final String ID_ELEMENT = "id";
    @NonNls
    private static final String NAME_ELEMENT = "name";

    @NonNls
    private static final String PLUGIN_VERSION_ELEMENT = "version";
    @NonNls
    private static final String IDEA_VERSION_ELEMENT = "idea-version";
    @NonNls
    private static final String SINCE_BUILD_ATTRIBUTE = "since-build";
    @NonNls
    private static final String UNTIL_BUILD_ATTRIBUTE = "until-build";

    private Document _document;
    private BuildNumber _buildNumber;
    private static final BuildNumber BUILD00 = BuildNumber.fromString("0.0");

    public PluginRepositoryManager(BuildNumber inBuildNumber) throws IOException, JDOMException {
        _buildNumber = inBuildNumber;
        _document = fetchDocument(INTELLIJ_NET_PLUGINS_LIST + inBuildNumber);
    }

    private Document fetchDocument(@NotNull String inPluginRepositoryUrl) throws IOException, JDOMException {
        Document document;

        HttpConfigurable.getInstance().prepareURL(inPluginRepositoryUrl);
        URL url = new URL(inPluginRepositoryUrl);
        URLConnection connection = url.openConnection();
        connection.setConnectTimeout(15 * 1000);
        connection.setReadTimeout(15 * 1000);
        connection.connect();

        InputStreamReader reader = null;
        try {
            InputStream is = connection.getInputStream();

            //boaz was:
            /*
            reader = new InputStreamReader(is, JDOMUtil.ENCODING);

            // create a SAX builder just like IDEA does, but don't use their JDOMUtil#loadDocument(InputStream) since this call seems not
            // thread-safe and triggers exceptions of type 'org.jdom.input.JDOMParseException: Error in building: Ill-formed XML document'
            SAXBuilder saxBuilder = new SAXBuilder();
            saxBuilder.setEntityResolver(new EntityResolver() {
                public InputSource resolveEntity(String inPublicId, String inSystemId) throws SAXException, IOException {
                    return new InputSource(new CharArrayReader(ArrayUtil.EMPTY_CHAR_ARRAY));
                }
            });

            document = saxBuilder.build(reader);
            */
            //boaz now:
            document = JDOMUtil.loadDocument(is);
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ioe) {
                    // nothing we can do at this point
                }
            }

            if (connection instanceof HttpURLConnection) {
                ((HttpURLConnection) connection).disconnect();
            }
        }

        if (LOG.isDebugEnabled()) {
            String content = JDOMUtil.writeDocument(document, "\n");
            LOG.debug("Downloaded plugin list: " + content);
        }

        return document;
    }

    @SuppressWarnings({"AnalyzingLoggingWithoutLogLevelCheck"})
    @NotNull
    public PluginRepositoryQueryResult isNewVersionAvailable(@NotNull IdeaPluginDescriptor inCurrentPluginDescriptor) {

        PluginId pluginId = inCurrentPluginDescriptor.getPluginId();
        if (pluginId == null) {
            return PluginRepositoryQueryResult.noNewVersion(inCurrentPluginDescriptor);
        }

        // in test mode, always signal availability of new version for certain plugins
        if (Boolean.valueOf(System.getProperty(TEST_MODE_PROPERTY_KEY))) {
            @NonNls String inTestPlugin1 = "SimpleIntentions";
            @NonNls String inTestPlugin2 = "SimpleActions";
            if (pluginId.getIdString().equals(inTestPlugin1) || pluginId.getIdString().equals(inTestPlugin2)) {
                return PluginRepositoryQueryResult.newVersion(inCurrentPluginDescriptor, "123.dummy");
            }
        }

        LOG.debug("Try to find repository entry in build range for plugin with id " + pluginId);
        Element element = findRepositoryEntry(pluginId, _buildNumber);
        if (element == null) {
            LOG.debug("No repository entry found in build range for plugin with id " + pluginId);
            return PluginRepositoryQueryResult.noNewVersion(inCurrentPluginDescriptor);
        }
        LOG.debug("Found repository entry in build range for plugin with id " + pluginId);

        if (LOG.isDebugEnabled()) {
            try {
                StringWriter stringWriter = new StringWriter();
                JDOMUtil.writeElement(element, stringWriter, "\n");
                String content = stringWriter.toString();
                LOG.debug("Element content of repository entry for plugin with id " + pluginId + ":\n" + content);
            } catch (IOException ioe) {
                LOG.info("Could not print dom element of repository entry for plugin with id " + pluginId, ioe);
            }
        }

        String pluginVersion = inCurrentPluginDescriptor.getVersion();
        if (pluginVersionHasNewer(pluginVersion, element)) {
            String newVersion = getPluginVersion(element);
            LOG.debug("New version  >" + newVersion + "< found for plugin with id " + pluginId);
            return PluginRepositoryQueryResult.newVersion(inCurrentPluginDescriptor, newVersion);
        }

        LOG.debug("No new version found for plugin with id " + pluginId);
        return PluginRepositoryQueryResult.noNewVersion(inCurrentPluginDescriptor);
    }

    @SuppressWarnings({"AnalyzingLoggingWithoutLogLevelCheck"})
    @Nullable
    private Element findRepositoryEntry(@NotNull PluginId inPluginId, BuildNumber inBuildNumber) {
        Element root = _document.getRootElement();
        Iterator pluginsIterator = root.getDescendants(new ElementFilter(IDEA_PLUGIN_ELEMENT));
        while (pluginsIterator.hasNext()) {
            Element ideaPluginElement = (Element) pluginsIterator.next();
            if (matchesPluginId(inPluginId, ideaPluginElement)) {
                LOG.debug("Repository entry's id/name matches plugin with id " + inPluginId);
                if (matchesIdeaVersion(inBuildNumber, ideaPluginElement)) {
                    LOG.debug("Repository entry's idea-version matches plugin with id " + inPluginId);
                    return ideaPluginElement;
                }
            }
        }

        return null;
    }

    private boolean matchesPluginId(@NotNull PluginId inPluginId, Element inIdeaPluginElement) {
        String id = inIdeaPluginElement.getChildTextTrim(ID_ELEMENT);
        String name = inIdeaPluginElement.getChildTextTrim(NAME_ELEMENT);

        if (id != null) {
            if (PluginId.getId(id).equals(inPluginId)) {
                return true;
            }
        } else if (name != null) {
            if (PluginId.getId(name).equals(inPluginId)) {
                return true;
            }
        }
        return false;
    }

    private boolean matchesIdeaVersion(BuildNumber inBuildNumber, @NotNull Element inIdeaPluginElement) {
        @Nullable Element ideaVersionElement = inIdeaPluginElement.getChild(IDEA_VERSION_ELEMENT);

        if (ideaVersionElement != null) {
            String sinceBuild = ideaVersionElement.getAttributeValue(SINCE_BUILD_ATTRIBUTE);
            String untilBuild = ideaVersionElement.getAttributeValue(UNTIL_BUILD_ATTRIBUTE);

            // check 'since' constraint
            if (!StringUtil.isEmptyOrSpaces(sinceBuild)) {
                try {
                    BuildNumber since = BuildNumber.fromString(sinceBuild);
                    //Integer since = new Integer(sinceBuild);
                    if (inBuildNumber.compareTo(since) < 0) {
                    //if (inBuildNumber < since) {
                        return false;
                    }
                } catch (Throwable nfe) {
                    return false;
                }
            }

            // check 'until' constraint
            if (!StringUtil.isEmptyOrSpaces(untilBuild)) {
                try {
                    BuildNumber until = BuildNumber.fromString(untilBuild);

                    // until is 0.0 make it a match
                    if (until.compareTo(BUILD00) == 0) {
                        return true;
                    }

                    //Integer until = new Integer(untilBuild);
                    if (inBuildNumber.compareTo(until) > 0) {
                        return false;
                    }
                } catch (NumberFormatException nfe) {
                    return false;
                }
            }

            // 'since' and 'until' match, if specified
            return true;
        }

        // no version element defined in XML definition, make it a match
        return true;
    }

    private boolean pluginVersionHasNewer(@Nullable String inVersion, @NotNull Element inIdeaPluginElement) {
        @Nullable String pluginVersion = getPluginVersion(inIdeaPluginElement);

        //boaz:
        return StringUtil.compareVersionNumbers(pluginVersion, inVersion) > 0;

        //return !ObjectUtils.equals(inVersion, pluginVersion);
    }

    @Nullable
    private String getPluginVersion(@NotNull Element inIdeaPluginElement) {
        @Nullable String pluginVersion = inIdeaPluginElement.getChildTextTrim(PLUGIN_VERSION_ELEMENT);
        return pluginVersion;
    }

    /**
     * Return null if failed to prepare
     */
    @Nullable
    public static PluginDownloader downloadPluginFromServer(@NotNull String inPluginId,
                                                            String pluginName,
                                                            String newVersion) throws IOException {
        File localPluginArchive = FileUtil.createTempFile("temp_" + inPluginId + "_", ".plugin");

        @NonNls String pluginIdToken = "PLUGIN_ID";
        @NonNls String buildToken = "BUILD";


        @NonNls final String url =
            RepositoryHelper.DOWNLOAD_URL + URLEncoder.encode(inPluginId, "UTF8") + "&build=" +
            ApplicationInfo.getInstance().getBuild().asString();
        // there is problem in PluginDownloader, fail to pass plugin name will cause
        // the loader to create invalid file name
        final PluginDownloader downloader = new PluginDownloader(inPluginId, url, newVersion, null, pluginName);
        if (downloader.prepareToInstall()) {
            return downloader;
            //downloader.install();
        } else {
            return null;
        }


        /*
        String pluginUrl = INTELLIJ_NET_PLUGIN_DOWNLOAD.replaceAll(pluginIdToken, inPluginId).
                replaceAll(buildToken, String.valueOf(inBuildNumber));

        if (LOG.isDebugEnabled()) {
            LOG.debug("Downloading plugin archive from: " + pluginUrl);
        }

        HttpConfigurable.getInstance().prepareURL(pluginUrl);
        URL url = new URL(pluginUrl);
        URLConnection connection = url.openConnection();
        connection.setConnectTimeout(15 * 1000);
        connection.setReadTimeout(15 * 1000);
        connection.connect();

        InputStream inputStream = null;
        OutputStream outputStream = null;
        try {
            inputStream = connection.getInputStream();
            outputStream = new FileOutputStream(localPluginArchive);
            StreamUtil.copyStreamContent(inputStream, outputStream);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException ioe) {
                    // nothing we can do at this point
                }
            }

            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException ioe) {
                    // nothing we can do at this point
                }
            }

            if (connection instanceof HttpURLConnection) {
                ((HttpURLConnection) connection).disconnect();
            }
        }

        return new Object[]{localPluginArchive, connection.getURL().toString()};
        */
    }

    public static class PluginRepositoryQueryResult {
        @NotNull
        private IdeaPluginDescriptor _currentDescriptor;
        private boolean _newVersionAvailable;
        @Nullable
        @NonNls
        private String _newVersion;

        public static PluginRepositoryQueryResult noNewVersion(@NotNull IdeaPluginDescriptor inCurrentDescriptor) {
            return new PluginRepositoryQueryResult(inCurrentDescriptor, false, null);
        }

        public static PluginRepositoryQueryResult newVersion(@NotNull IdeaPluginDescriptor inCurrentDescriptor,
                                                             @Nullable @NonNls String inNewVersion) {
            return new PluginRepositoryQueryResult(inCurrentDescriptor, true, inNewVersion);
        }

        private PluginRepositoryQueryResult(@NotNull IdeaPluginDescriptor inCurrentDescriptor, boolean inNewVersionAvailable,
                                            @Nullable @NonNls String inNewVersion) {
            _currentDescriptor = inCurrentDescriptor;
            _newVersionAvailable = inNewVersionAvailable;
            _newVersion = inNewVersion;
        }

        @NotNull
        public IdeaPluginDescriptor getCurrentDescriptor() {
            return _currentDescriptor;
        }

        public boolean isNewVersionAvailable() {
            return _newVersionAvailable;
        }

        @Nullable
        @NonNls
        public String getNewVersion() {
            return _newVersion;
        }
    }
}
