/*
 * Copyright 2007 Etienne Studer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nu.studer.idea.notifier;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.*;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.options.Configurable;
import com.intellij.openapi.options.ConfigurationException;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.util.DefaultJDOMExternalizer;
import com.intellij.openapi.util.InvalidDataException;
import com.intellij.openapi.util.NamedJDOMExternalizable;
import com.intellij.openapi.util.WriteExternalException;

import org.jdom.Element;

/**
 * This class allows for configuration of the plugin update notifier.
 *
 * @author <a href="mailto:intellij@studer.nu">Etienne Studer</a>, Nov 7, 2006
 */
public class PluginUpdateNotifierConfigurable implements ApplicationComponent, NamedJDOMExternalizable, Configurable {
    @NotNull
    @Nls
    public static final String HOURLY = PluginUpdateNotifierBundle.message("hourly");
    @NotNull
    @Nls
    public static final String DAILY = PluginUpdateNotifierBundle.message("daily");
    @NotNull
    @Nls
    public static final String WEEKLY = PluginUpdateNotifierBundle.message("weekly");

    @NotNull
    public static final String[] INTERVALS = new String[]{HOURLY, DAILY, WEEKLY};

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public String UPDATE_INTERVAL;// persisted setting

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public boolean CHECK_UPDATES;// persisted setting

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public boolean IGNORE_DISABLED_PLUGINS;// persisted setting

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public boolean IGNORE_BUNDLED_PLUGINS;// persisted setting

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public long LAST_EXECUTION;// persisted setting

    @SuppressWarnings({"AnalyzingVariableNaming"})
    public long NEXT_EXECUTION;// persisted setting

    private PluginUpdateNotifierConfigurationPane _configurationPane;

    public static PluginUpdateNotifierConfigurable getInstance() {
        return ApplicationManager.getApplication().getComponent(PluginUpdateNotifierConfigurable.class);
    }

    public PluginUpdateNotifierConfigurable() {
        UPDATE_INTERVAL = DAILY;
        CHECK_UPDATES = true;
        IGNORE_DISABLED_PLUGINS = false;
        IGNORE_BUNDLED_PLUGINS = true;
        LAST_EXECUTION = 0;
        NEXT_EXECUTION = System.currentTimeMillis();
    }

    @NotNull
    public String getComponentName() {
        return "PluginUpdateNotifier.PluginUpdateNotifierConfigurable";
    }

    public void initComponent() {
    }

    public void disposeComponent() {
    }

    public void readExternal(Element inElement) throws InvalidDataException {
        DefaultJDOMExternalizer.readExternal(this, inElement);
    }

    public void writeExternal(Element inElement) throws WriteExternalException {
        DefaultJDOMExternalizer.writeExternal(this, inElement);
    }

    @NonNls
    public String getExternalFileName() {
        return "pluginUpdateNotifier";
    }

    @Nls
    public String getDisplayName() {
        return PluginUpdateNotifierBundle.message("plugin.notifier.title");
    }

    public Icon getIcon() {
        return null;
    }

    @Nullable
    @NonNls
    public String getHelpTopic() {
        return null;
    }

    public JComponent createComponent() {
        if (_configurationPane == null) {
            _configurationPane = new PluginUpdateNotifierConfigurationPane(INTERVALS);
            _configurationPane.getCheckNowButton().addActionListener(new CheckNowActionListener());
            _configurationPane.getRootComponent().addAncestorListener(new ConfigurationPaneAncestorListener());
        }
        return _configurationPane.getRootComponent();
    }

    public boolean isModified() {
        return _configurationPane != null && _configurationPane.isModified(this);
    }

    public void apply() throws ConfigurationException {
        if (_configurationPane != null) {
            _configurationPane.getData(this);

            // upon changing the update interval, we must
            // 1) recalculate the calculated fields (next execution time)
            // 2) refresh the configuration view with the new calculated field values
            // 3) notify the background thread about the configuration change
            adjustNextExecutionTime(System.currentTimeMillis());
            _configurationPane.setShowTime(HOURLY.equals(UPDATE_INTERVAL));
            _configurationPane.setData(this);
            PluginUpdateNotifierBackgroundComponent.getInstance().applyConfigurationChange();
        }
    }

    public void reset() {
        if (_configurationPane != null) {
            _configurationPane.setShowTime(HOURLY.equals(UPDATE_INTERVAL));
            _configurationPane.setData(this);
        }
    }

    public void disposeUIResources() {
        _configurationPane = null;
    }

    public void updateNextExecutionTime(boolean inSuccessful) {
        if (inSuccessful) {
            LAST_EXECUTION = System.currentTimeMillis();

            // increase next execution time by update interval if last plugin manager query has been successful
            adjustNextExecutionTime(LAST_EXECUTION);
        } else {
            // retry execution in two hours if last plugin manager query has NOT been successful
            NEXT_EXECUTION = System.currentTimeMillis() + PluginUpdateNotifierUtil.TWO_HOURS;
        }
    }

    private void adjustNextExecutionTime(long inBaseline) {
        if (HOURLY.equals(UPDATE_INTERVAL)) {
            NEXT_EXECUTION = inBaseline + PluginUpdateNotifierUtil.ONE_HOUR;
        } else if (DAILY.equals(UPDATE_INTERVAL)) {
            NEXT_EXECUTION = inBaseline + PluginUpdateNotifierUtil.ONE_DAY;
        } else if (WEEKLY.equals(UPDATE_INTERVAL)) {
            NEXT_EXECUTION = inBaseline + PluginUpdateNotifierUtil.ONE_WEEK;
        }
    }

    /**
     * This listener triggers an immediate check for outofdate plugins.
     */
    private static class CheckNowActionListener implements ActionListener {
        public void actionPerformed(ActionEvent inEvent) {
            // query plugin repository manager in background thread
            Runnable runnable = new Runnable() {
                public void run() {
                    ProgressIndicator indicator = ProgressManager.getInstance().getProgressIndicator();
                    indicator.setText(PluginUpdateNotifierBundle.message("progress.dialog.text"));
                    indicator.setIndeterminate(true);

                    PluginUpdateNotifierBackgroundComponent.getInstance().queryPluginRepositoryManager();
                }
            };
            ProgressManager progressManager = ProgressManager.getInstance();
            progressManager.runProcessWithProgressSynchronously(runnable, PluginUpdateNotifierBundle.message("progress.dialog.title"), false, null);

            // show info dialog once the updates are available
            PluginUpdateNotifierDialog dialog = new PluginUpdateNotifierDialog(null);
            dialog.show();
        }
    }

    /**
     * This listener packs the specified component's ancestor window when the component is added to that window.
     */
    private static class ConfigurationPaneAncestorListener implements AncestorListener {
        public void ancestorAdded(AncestorEvent inEvent) {
            Component source = (Component) inEvent.getSource();
            if (source != null) {
                Window window = SwingUtilities.getWindowAncestor(source);
                if (window != null) {
                    window.pack();
                }
            }
        }

        public void ancestorMoved(AncestorEvent inEvent) {
        }

        public void ancestorRemoved(AncestorEvent inEvent) {
        }
    }
}
