/*
 * Copyright 2007 Etienne Studer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nu.studer.idea.notifier;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.util.List;
import javax.swing.*;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import com.intellij.ide.plugins.PluginManagerConfigurable;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.options.ShowSettingsUtil;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.ui.Messages;
import com.intellij.ui.components.labels.LinkLabel;
import com.intellij.ui.components.labels.LinkListener;

/**
 * This class defines the dialog that provides a summary of the outofdate plugins.
 *
 * @author <a href="mailto:intellij@studer.nu">Etienne Studer</a>, Nov 13, 2006
 */
public class PluginUpdateNotifierDialog extends DialogWrapper {
    private Project _project;
    private LinkLabel _pluginManagerLinkLabel;
    private JTextArea _messageTextArea;
    private Action _updateAction;

    public PluginUpdateNotifierDialog(Project inProject) {
        super(inProject, true);
        _project = inProject;

        setTitle(PluginUpdateNotifierBundle.message("plugin.notifier.title"));

        _pluginManagerLinkLabel = new LinkLabel(PluginUpdateNotifierBundle.message("plugin.manager.link"), null, new LinkListener() {
            public void linkSelected(LinkLabel inSource, Object inData) {
                // close the info dialog
                close(OK_EXIT_CODE);

                // show the plugin manager dialog
                PluginManagerConfigurable pluginManager = PluginManagerConfigurable.getInstance();
                ShowSettingsUtil.getInstance().editConfigurable(_project, pluginManager);
            }
        });

        _messageTextArea = new JTextArea();
        _messageTextArea.setOpaque(false);
        _messageTextArea.setEditable(false);
        _messageTextArea.setFont(_pluginManagerLinkLabel.getFont());

        _updateAction = new AbstractAction(PluginUpdateNotifierBundle.message("plugin.update.action.text")) {
            public void actionPerformed(ActionEvent inEvent) {
                downloadPluginUpdates();
            }
        };

        init();
    }

    @Nullable
    protected JComponent createCenterPanel() {
        JPanel panel = new JPanel(new BorderLayout(0, 10));
        panel.add(_messageTextArea, BorderLayout.CENTER);
        panel.add(_pluginManagerLinkLabel, BorderLayout.SOUTH);
        return panel;
    }

    protected Action[] createActions() {
        return new Action[]{getOKAction(), _updateAction};
    }

    public void show() {
        updateMessageTextArea();
        updateUpdateButton();
        super.show();
    }

    private void updateMessageTextArea() {
        StringBuffer finalMessage = new StringBuffer();

        final PluginUpdateNotifierBackgroundComponent.UpdateState updateState = getBackgroundComponent().getUpdateState();
        final List<PluginRepositoryManager.PluginRepositoryQueryResult> outOfDatePlugins = updateState.getOutOfDatePlugins();
        final Exception exception = updateState.getException();

        // append plugin state message, if needed
        if (outOfDatePlugins != null) {
            StringBuffer notification = PluginUpdateNotifierUtil.createNotificationMessage(outOfDatePlugins, true);
            finalMessage.append(notification);
        }

        // append error message, if needed
        if (exception != null) {
            if (finalMessage.length() > 0) {
                finalMessage.append("\n\n");
            }

            String errorMessage = PluginUpdateNotifierBundle.message("unable.to.retrieve.plugin.repos.info.verbose") + "\n" + exception.getMessage();
            finalMessage.append(errorMessage);
        }

        _messageTextArea.setText(finalMessage.toString());
    }

    private void updateUpdateButton() {
        final PluginUpdateNotifierBackgroundComponent.UpdateState updateState = getBackgroundComponent().getUpdateState();
        final List<PluginRepositoryManager.PluginRepositoryQueryResult> outOfDatePlugins = updateState.getOutOfDatePlugins();

        _updateAction.setEnabled(outOfDatePlugins != null && outOfDatePlugins.size() > 0);
    }

    private void downloadPluginUpdates() {
        final PluginUpdateNotifierBackgroundComponent.UpdateState updateState = getBackgroundComponent().getUpdateState();
        final List<PluginRepositoryManager.PluginRepositoryQueryResult> outOfDatePlugins = updateState.getOutOfDatePlugins();

        // just to be safe, should never be true
        if (outOfDatePlugins == null || outOfDatePlugins.size() == 0) {
            return;
        }

        PluginUpdateNotifierDownloadUtil pluginDownloader = new PluginUpdateNotifierDownloadUtil(_project);
        List<IOException> exceptions = pluginDownloader.downloadUpdates(outOfDatePlugins);
        _updateAction.setEnabled(exceptions.size() != 0); // disable action if everything went well

        if (exceptions.size() > 0) {
            StringBuffer message = concatenateMessages(exceptions);
            Messages.showErrorDialog(_project, PluginUpdateNotifierBundle.message("progress.update.dialog.failure.text.0", message),
                    PluginUpdateNotifierBundle.message("progress.update.dialog.failure.title"));
        }
    }

    @NotNull
    private StringBuffer concatenateMessages(@NotNull List<IOException> inExceptions) {
        StringBuffer result = new StringBuffer();
        for (IOException exception : inExceptions) {
            if (result.length() > 0) {
                result.append("\n");
            }
            result.append(exception.getMessage());
        }
        return result;
    }

    private static PluginUpdateNotifierBackgroundComponent getBackgroundComponent() {
        Application application = ApplicationManager.getApplication();
        PluginUpdateNotifierBackgroundComponent backgroundComponent = application.getComponent(PluginUpdateNotifierBackgroundComponent.class);
        return backgroundComponent;
    }
}
