/*
 * Copyright 2007 Etienne Studer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nu.studer.idea.notifier;

import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ex.ApplicationManagerEx;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.updateSettings.impl.PluginDownloader;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * This class contains utility methods related to downloading new plugin versions.
 *
 * @author <a href="mailto:estuder@navis.com">Etienne Studer</a>, Sep 12, 2007
 * @author KIR (basic action script functionality)
 */
public class PluginUpdateNotifierDownloadUtil {
    private static final Logger LOG = Logger.getInstance(PluginUpdateNotifierDownloadUtil.class.getName());

    private final Project _project;

    public PluginUpdateNotifierDownloadUtil(Project inProject) {
        _project = inProject;
    }

    @NotNull
    public List<IOException> downloadUpdates(@NotNull final List<PluginRepositoryManager.PluginRepositoryQueryResult> inPlugins) {
        final List<IOException> exceptions = new ArrayList<IOException>();

        // download plugin updates in background thread
        Runnable runnable = new Runnable() {
            public void run() {
                ProgressIndicator indicator = ProgressManager.getInstance().getProgressIndicator();
                indicator.setText(PluginUpdateNotifierBundle.message("progress.update.dialog.text"));
                indicator.setIndeterminate(true);

                List<IOException> downloadExceptions = internalDownloadUpdates(inPlugins);
                exceptions.addAll(downloadExceptions);
            }
        };
        ProgressManager progressManager = ProgressManager.getInstance();
        boolean successful = progressManager.runProcessWithProgressSynchronously(runnable,
                PluginUpdateNotifierBundle.message("progress.update.dialog.title"), false, _project);

        // in case the CANCEL button is shown on the progress dialog and the user presses it (we currently do not show the CANCEL button)
        if (!successful) {
            return Collections.emptyList();
        }

        // in case exceptions have occurred
        if (exceptions.size() > 0) {
            return exceptions;
        }

        // everything went well, prompt fur shutdown
        promptShutdownAndShutdown();
        return Collections.emptyList();
    }

    @NotNull
    private List<IOException> internalDownloadUpdates(List<PluginRepositoryManager.PluginRepositoryQueryResult> inPlugins) {
        List<IOException> exceptions = new ArrayList<IOException>();
        for (PluginRepositoryManager.PluginRepositoryQueryResult inPlugin : inPlugins) {
            // meta information
            IdeaPluginDescriptor currentDescriptor = inPlugin.getCurrentDescriptor();
            String pluginId = currentDescriptor.getPluginId().getIdString();
            //BuildNumber ideaBuildVersion = ApplicationInfo.getInstance().getBuild();

            // update progress dialog before starting the download process
            ProgressIndicator indicator = ProgressManager.getInstance().getProgressIndicator();
            indicator.setText(PluginUpdateNotifierBundle.message("progress.update.dialog.text.specific.0", pluginId));

            try {
                PluginDownloader downloader = PluginRepositoryManager.downloadPluginFromServer(
                    pluginId,
                    currentDescriptor.getName(),
                    inPlugin.getNewVersion());

                if (downloader == null) {
                    throw new IOException("Uanble to download '" + pluginId + "' from unknown reason");
                }


                //File localArchiveFile = (File) result[0];
                //String serverPath = (String) result[1];

                if (LOG.isDebugEnabled()) {
                    LOG.debug("Updating plugin with id " + pluginId);
                    LOG.debug("Updated local plugin archive file: " + downloader.getFileName());
                    //LOG.debug("Downloaded from server path: " + serverPath);
                }

                downloader.install();

                // add startup actions
                //addActions(inPlugin.getCurrentDescriptor(), localArchiveFile, serverPath);
            } catch (IOException ioe) {
                LOG.info("Error downloading plugin update for plugin >" + pluginId + "<", ioe);
                exceptions.add(ioe);
            }
        }
        return exceptions;
    }

    /*
    private void addActions(IdeaPluginDescriptor inPluginDescriptor, File inLocalArchiveFile, String inServerPath) throws IOException {
        PluginId id = inPluginDescriptor.getPluginId();
        if (PluginManager.isPluginInstalled(id)) {
            // store old plugins file
            File oldFile = inPluginDescriptor.getPath();
            StartupActionScriptManager.ActionCommand deleteOld = new StartupActionScriptManager.DeleteCommand(oldFile);
            StartupActionScriptManager.addActionCommand(deleteOld);
        }

        //noinspection HardCodedStringLiteral
        boolean isJarFile = inServerPath.endsWith(".jar");
        if (isJarFile) {
            // add command to copy file to the IDEA/plugins path
            String fileName = new File(inServerPath).getName();
            File newFile = new File(PathManager.getPluginsPath() + File.separator + fileName);
            StartupActionScriptManager.ActionCommand copyPlugin = new StartupActionScriptManager.CopyCommand(inLocalArchiveFile, newFile);
            StartupActionScriptManager.addActionCommand(copyPlugin);
        } else {
            // add command to unzip file to the IDEA/plugins path
            String unzipPath;
            if (ZipUtil.isZipContainsFolder(inLocalArchiveFile)) {
                unzipPath = PathManager.getPluginsPath();
            } else {
                String dirName = inPluginDescriptor.getName();
                unzipPath = PathManager.getPluginsPath() + File.separator + dirName;
            }

            File newFile = new File(unzipPath);
            StartupActionScriptManager.ActionCommand unzip = new StartupActionScriptManager.UnzipCommand(inLocalArchiveFile, newFile);
            StartupActionScriptManager.addActionCommand(unzip);
        }

        // add command to remove temp plugin file
        StartupActionScriptManager.ActionCommand deleteTemp = new StartupActionScriptManager.DeleteCommand(inLocalArchiveFile);
        StartupActionScriptManager.addActionCommand(deleteTemp);
    }
    */

    private void promptShutdownAndShutdown() {
        ApplicationManager.getApplication().invokeLater(new Runnable() {
            public void run() {
                int choice = Messages.showDialog(PluginUpdateNotifierBundle.message("progress.shutdown.dialog.text"),
                        PluginUpdateNotifierBundle.message("progress.shutdown.dialog.title"),
                        new String[]{PluginUpdateNotifierBundle.message("progress.shutdown.dialog.button.shutdown.text"),
                            PluginUpdateNotifierBundle.message("progress.shutdown.dialog.button.restart.text"),
                                PluginUpdateNotifierBundle.message("progress.shutdown.dialog.button.cancel.text")},
                        0, Messages.getQuestionIcon());

                if (choice == 0) {
                    ApplicationManagerEx.getApplicationEx().exit(true);
                } else if (choice == 1) {
                    ApplicationManagerEx.getApplicationEx().restart();
                }
            }
        });
    }

    @SuppressWarnings({"HardCodedStringLiteral"})
    public static void main(String[] inArgs) {
        File file = new File("http://plugins.intellij.net/files/GenerateToString_3055.jar");
        System.out.println("file.getName() = " + file.getName());
    }
}
